from redisx import define
from typing import (Sequence, Tuple, Union)
from functools import reduce
from redis.typing import (CommandsProtocol)

VectorType = Sequence[Union[int, float]]


class TextVectorEncoder:
    SEP = bytes(",", "ascii")
    BITS = ("0", "1")

    @classmethod
    def vector_str(cls, vector: VectorType) -> bytes:
        """
        @return e.g: "0.1,0.0,0.2"
        """
        s = ",".join(["%f" % x for x in vector])
        return bytes(s, encoding="ascii")

    @classmethod
    def encode_with_square_brackets(cls, vector: VectorType, is_binary=False) -> bytes:
        s = ""
        if is_binary:
            s = "[" + ",".join([cls.BITS[x] for x in vector]) + "]"
        else:
            s = "[" + ",".join(["%f" % x for x in vector]) + "]"
        return bytes(s, encoding="ascii")  # ascii is enough

    @classmethod
    def decode_with_square_brackets(cls, buf: bytes) -> Tuple[float]:
        if buf[0] != ord("[") or buf[-1] != ord("]"):
            raise ValueError("invalid text vector value")
        is_int = True
        components = buf[1:-1].split(cls.SEP)
        for x in components:
            if not x.isdigit():
                is_int = False

        if is_int:
            return tuple(int(x) for x in components)
        return tuple(float(x) for x in components)


class UsearchVectorCommands(CommandsProtocol):

    def create_index(
        self,
        name: str,
        dim: int,
        m: int = 10,
        efcon: int = 128,
        metric: str = define.DistanceMetric.InnerProduct,
        quantization: str = define.UsearchQuantizationType.F32,
        **kwargs
    ):
        """
        create a index
        cmd eg: usearch.index.create idx0 dim 3 m 10 efcon 12 metric ip quantization f32
        """
        params = reduce(lambda x, y: x + y, kwargs.items(), ())
        return self.execute_command(
            define.CmdName.USEARCH_CREATE_INDEX,
            name,
            "dim",
            dim,
            "m",
            m,
            "efcon",
            efcon,
            "metric",
            metric,
            "quantization",
            quantization,
            *params
        )

    def get_index(self, name: str):
        """
        get the infomation of an index
        """
        return self.execute_command(
            define.CmdName.USEARCH_GET_INDEX,
            name
        )

    def del_index(self, name: str):
        """
        delete an index and all its data
        """
        return self.execute_command(
            define.CmdName.USEARCH_DEL_INDEX,
            name
        )

    def add_vector(
            self,
            index_name: str,
            name: str,
            vector: Union[VectorType, str],
    ):
        """
        add vector to index by name
        """
        if not isinstance(vector, str):
            vector = TextVectorEncoder.vector_str(vector)
        return self.execute_command(
            define.CmdName.USEARCH_ADD_NODE,
            index_name,
            name,
            vector
        )

    def add_vector_id(
            self,
            index_name: str,
            id: int,
            vector: Union[VectorType, str],
    ):
        """
        add vector to index by id
        """
        if not isinstance(vector, str):
            vector = TextVectorEncoder.vector_str(vector)
        return self.execute_command(
            define.CmdName.USEARCH_ADD_ID_NODE,
            index_name,
            id,
            vector
        )

    def get_vector(self, index_name: str, name: str):
        """
        get vector from index by name
        """
        return self.execute_command(
            define.CmdName.USEARCH_GET_NODE,
            index_name,
            name
        )

    def get_vector_id(self, index_name: str, id: int):
        """
        get vector from index by id
        """
        return self.execute_command(
            define.CmdName.USEARCH_GET_ID_NODE,
            index_name,
            id
        )

    def del_vector(self, index_name: str, name: str):
        """
        del vector from index by name
        """
        return self.execute_command(
            define.CmdName.USEARCH_DEL_NODE,
            index_name,
            name
        )

    def del_vector_id(self, index_name: str, id: int):
        """
        del vector from index by id
        """
        return self.execute_command(
            define.CmdName.USEARCH_DEL_ID_NODE,
            index_name,
            id
        )

    def kann_search(
        self,
        index_name: str,
        k: int,
        query_vector: Union[VectorType, str],
        **kwargs
    ):
        """
        KANN search from index by query vector
        cmd eg: usearch.search.kann idx0 6 0.0,0.0,0.0 "" ef_search 10
        """
        params = reduce(lambda x, y: x + y, kwargs.items(), ())
        if not isinstance(query_vector, str):
            query_vector = TextVectorEncoder.vector_str(query_vector)
        return self.execute_command(
            define.CmdName.USEARCH_SEARCH_KANN,
            index_name,
            k,
            query_vector,
            *params
        )



from typing import Dict, Union

from redisx import ann_usearch
from redisx import define

from redis._parsers.helpers import (
    bool_ok,
    pairs_to_dict,
    parse_list_of_dicts,
)
from redis import Redis
from redis.asyncio import Redis as AsyncRedis


# just a commands class set for extend
class RedisXCommands(
    ann_usearch.UsearchVectorCommands,
):
    pass


def int_or_none(response):
    if response is None:
        return None
    return int(response)


def parse_usearch_get_index_result(resp) -> Union[Dict, None]:
    if len(resp) == 0:
        return None
    return pairs_to_dict(resp, decode_keys=True, decode_string_values=True)


def parse_usearch_get_node_result(resp) -> Union[Dict, None]:
    if len(resp) == 0:
        return None
    return pairs_to_dict(resp, decode_keys=True, decode_string_values=True)


def parse_usearch_kann_search_result(resp) -> Union[Dict, None]:
    print("====== {}\n".format(resp))
    if len(resp) == 1:
        return {"size": int(resp[0]), "vals": None}
    if len(resp) > 1:
        vals = []
        for item in resp[1:]:
            vals.append(pairs_to_dict(
                item, decode_keys=True, decode_string_values=True))
        return {"size": int(resp[0]), "vals": vals}
    return None


REDISX_RESPONSE_CALLBACKS = {
    # RedisXANN Usearch Vector
    define.CmdName.USEARCH_CREATE_INDEX: bool_ok,
    define.CmdName.USEARCH_GET_INDEX: parse_usearch_get_index_result,
    define.CmdName.USEARCH_DEL_INDEX: int_or_none,
    define.CmdName.USEARCH_ADD_NODE: bool_ok,
    define.CmdName.USEARCH_GET_NODE: parse_usearch_get_node_result,
    define.CmdName.USEARCH_DEL_NODE: int_or_none,
    define.CmdName.USEARCH_ADD_ID_NODE: bool_ok,
    define.CmdName.USEARCH_GET_ID_NODE: parse_usearch_get_node_result,
    define.CmdName.USEARCH_DEL_ID_NODE: int_or_none,
    define.CmdName.USEARCH_SEARCH_KANN: parse_usearch_kann_search_result,
}


def set_response_callback(redis: Union[Redis, AsyncRedis]):
    for cmd, cb in REDISX_RESPONSE_CALLBACKS.items():
        redis.set_response_callback(cmd, cb)



class DistanceMetric:
    Euclidean = "L2"  # an alias to L2
    L2 = "L2"
    InnerProduct = "IP"
    Jaccard = "JACCARD"
    Cosine = "COS"
    L2sq = "L2SQ",
    Pearson = "PEARSON",
    Haversine = "HAVERSINE",
    Hamming = "HAMMING",
    Tanimoto = "TANIMOTO",
    Sorensen = "SORENSEN",


# downcast
class UsearchQuantizationType:
    F64 = "F64"
    F32 = "F32"
    F16 = "F16"
    I8 = "I8"
    B1 = "B1"


class CmdName:
    USEARCH_CREATE_INDEX = "USEARCH.INDEX.CREATE"
    USEARCH_GET_INDEX = "USEARCH.INDEX.GET"
    USEARCH_DEL_INDEX = "USEARCH.INDEX.DEL"
    USEARCH_ADD_NODE = "USEARCH.NODE.ADD"
    USEARCH_GET_NODE = "USEARCH.NODE.GET"
    USEARCH_DEL_NODE = "USEARCH.NODE.DEL"
    USEARCH_ADD_ID_NODE = "USEARCH.NODE.ADD_ID"
    USEARCH_GET_ID_NODE = "USEARCH.NODE.GET_ID"
    USEARCH_DEL_ID_NODE = "USEARCH.NODE.DEL_ID"
    USEARCH_SEARCH_KANN = "USEARCH.SEARCH.KANN"



import unittest

from redis import ResponseError
from conf_cases import get_client


cli = get_client()


class IndexCommandsTest(unittest.TestCase):
    def __init__(self, methodName="runTest"):
        super().__init__(methodName=methodName)
        self.dim = 5
        self.m = 32
        self.efconf = 200
        self.other_params = {}

    # unittest sort alphabetically
    def test_0_delete_noexist_index(self):
        with self.assertRaises(ResponseError) as re:
            cli.del_index("test")
        self.assertEqual(
            str(re.exception),
            'Index: usearch.test does not exist',
        )

    def test_1_get_nonexist_index(self):
        with self.assertRaises(ResponseError) as re:
            cli.get_index("test")
        self.assertEqual(
            str(re.exception),
            'Index: usearch.test does not exist',
        )

    def test_2_create_index(self):
        res = cli.create_index(
            "test",
            self.dim,
            self.m,
            efcon=self.efconf,
            **self.other_params
        )
        self.assertEqual(res, True)

    def test_3_get_exist_index(self):
        res = cli.get_index("test")
        self.assertEqual(res["dimensions"], self.dim)
        self.assertEqual(res["connectivity"], self.m)
        self.assertEqual(res["expansion_add"], self.efconf)

    def test_4_delete_exist_index(self):
        res = cli.del_index("test")
        self.assertEqual(res, 1)


if __name__ == "__main__":
    unittest.main()
    cli.close()



import pytest
from redis.exceptions import RedisClusterException
from redisx import Client, ClusterClient


class TestPipeline:
    def test_pipeline_is_true(self, uc: Client):
        with uc.pipeline() as pipe:
            assert pipe

    def test_deprecated(self, cc: ClusterClient):
        with pytest.raises(RedisClusterException):
            cc.pipeline(transaction=True)
        with pytest.raises(RedisClusterException):

            cc.pipeline(shard_hint=True)


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
from setuptools import setup

setup(
    name="redisx",
    description="Python client for RedisX",
    long_description=open("README.md").read().strip(),
    long_description_content_type="text/markdown",
    version="0.1.4",
    license="MIT",
    url="https://github.com/weedge/redisx-py",
    author="weedge",
    author_email="weege007@gmail.com",
    python_requires=">=3.10",
    packages=["redisx"],
    install_requires=["redis == 5.0.0"],
)



from redis import ConnectionPool, Redis
from redis import RedisCluster
from redis.exceptions import RedisClusterException

from redisx.commands import RedisXCommands, set_response_callback
from redisx.pipeline import Pipeline, ClusterPipeline


# wrap redis client with redisx cmd
class Client(Redis, RedisXCommands):
    @classmethod
    def from_url(cls, url: str, **kwargs):
        connection_pool = ConnectionPool.from_url(url, **kwargs)
        return cls(connection_pool=connection_pool)

    def __init__(
        self,
        host="localhost",
        port=6379,
        db=0,
        password=None,
        socket_timeout=None,
        socket_connect_timeout=None,
        socket_keepalive=None,
        socket_keepalive_options=None,
        connection_pool=None,
        unix_socket_path=None,
        encoding="utf-8",
        encoding_errors="strict",
        charset=None,
        errors=None,
        decode_responses=False,
        retry_on_timeout=False,
        retry_on_error=[],
        ssl=False,
        ssl_keyfile=None,
        ssl_certfile=None,
        ssl_cert_reqs="required",
        ssl_ca_certs=None,
        ssl_ca_path=None,
        ssl_ca_data=None,
        ssl_check_hostname=False,
        ssl_password=None,
        ssl_validate_ocsp=False,
        ssl_validate_ocsp_stapled=False,
        ssl_ocsp_context=None,
        ssl_ocsp_expected_cert=None,
        max_connections=None,
        single_connection_client=False,
        health_check_interval=0,
        client_name=None,
        username=None,
        retry=None,
        redis_connect_func=None,
    ):
        Redis.__init__(
            self,
            host=host,
            port=port,
            db=db,
            password=password,
            socket_timeout=socket_timeout,
            socket_connect_timeout=socket_connect_timeout,
            socket_keepalive=socket_keepalive,
            socket_keepalive_options=socket_keepalive_options,
            connection_pool=connection_pool,
            unix_socket_path=unix_socket_path,
            encoding=encoding,
            encoding_errors=encoding_errors,
            charset=charset,
            errors=errors,
            decode_responses=decode_responses,
            retry_on_timeout=retry_on_timeout,
            retry_on_error=retry_on_error,
            ssl=ssl,
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
            ssl_cert_reqs=ssl_cert_reqs,
            ssl_ca_certs=ssl_ca_certs,
            ssl_ca_path=ssl_ca_path,
            ssl_ca_data=ssl_ca_data,
            ssl_check_hostname=ssl_check_hostname,
            ssl_password=ssl_password,
            ssl_validate_ocsp=ssl_validate_ocsp,
            ssl_validate_ocsp_stapled=ssl_validate_ocsp_stapled,
            ssl_ocsp_context=ssl_ocsp_context,
            ssl_ocsp_expected_cert=ssl_ocsp_expected_cert,
            max_connections=max_connections,
            single_connection_client=single_connection_client,
            health_check_interval=health_check_interval,
            client_name=client_name,
            username=username,
            retry=retry,
            redis_connect_func=redis_connect_func,
        )
        set_response_callback(self)

    def pipeline(self, transaction=True, shard_hint=None):
        return Pipeline(
            self.connection_pool,
            self.response_callbacks,
            transaction,
            shard_hint,
        )


class ClusterClient(RedisCluster, RedisXCommands):
    @classmethod
    def from_url(cls, url, **kwargs):
        return cls(url=url, **kwargs)

    def __init__(
        self,
        host=None,
        port=6379,
        startup_nodes=None,
        cluster_error_retry_attempts=3,
        require_full_coverage=False,
        reinitialize_steps=10,
        read_from_replicas=False,
        url=None,
        **kwargs,
    ):
        RedisCluster.__init__(
            self,
            host=host,
            port=port,
            startup_nodes=startup_nodes,
            cluster_error_retry_attempts=cluster_error_retry_attempts,
            require_full_coverage=require_full_coverage,
            reinitialize_steps=reinitialize_steps,
            read_from_replicas=read_from_replicas,
            url=url,
            **kwargs,
        )
        set_response_callback(self)

    def pipeline(self, transaction=None, shard_hint=None):
        if shard_hint:
            raise RedisClusterException(
                "shard_hint is deprecated in cluster mode")

        if transaction:
            raise RedisClusterException(
                "transaction is deprecated in cluster mode")

        return ClusterPipeline(
            nodes_manager=self.nodes_manager,
            commands_parser=self.commands_parser,
            startup_nodes=self.nodes_manager.startup_nodes,
            result_callbacks=self.result_callbacks,
            cluster_response_callbacks=self.cluster_response_callbacks,
            cluster_error_retry_attempts=self.cluster_error_retry_attempts,
            read_from_replicas=self.read_from_replicas,
            reinitialize_steps=self.reinitialize_steps,
        )
